<?php
/**
 * Copyright 2017 Lengow SAS
 *
 * NOTICE OF LICENSE
 *
 * This source file is subject to the Open Software License (OSL 3.0)
 * that is bundled with this package in the file LICENSE.txt.
 * It is also available through the world-wide-web at this URL:
 * http://opensource.org/licenses/osl-3.0.php
 *
 * @category    Lengow
 * @package     Lengow_Connector
 * @subpackage  Model
 * @author      Team module <team-module@lengow.com>
 * @copyright   2017 Lengow SAS
 * @license     http://opensource.org/licenses/osl-3.0.php  Open Software License (OSL 3.0)
 */

namespace Lengow\Connector\Model;

use Magento\Catalog\Model\Product\Attribute\Source\Status as ProductStatus;
use Magento\Catalog\Model\ResourceModel\Product\Collection as ProductCollection;
use Magento\Catalog\Model\ResourceModel\Product\CollectionFactory as ProductCollectionFactory;
use Magento\CatalogInventory\Model\Configuration as CatalogInventoryConfiguration;
use Magento\Framework\App\Config\ScopeConfigInterface;
use Magento\Framework\Json\Helper\Data as JsonHelper;
use Magento\Framework\Stdlib\DateTime\DateTime;
use Magento\Store\Model\ScopeInterface;
use Magento\Store\Model\StoreManagerInterface;
use Magento\Store\Model\Store\Interceptor as StoreInterceptor;
use Magento\Store\Model\WebsiteFactory;
use Lengow\Connector\Helper\Config as ConfigHelper;
use Lengow\Connector\Helper\Data as DataHelper;
use Lengow\Connector\Model\Exception as LengowException;
use Lengow\Connector\Model\Export\Feed as LengowFeed;
use Lengow\Connector\Model\Export\FeedFactory as LengowFeedFactory;
use Lengow\Connector\Model\Export\ProductFactory as LengowProductFactory;
use Lengow\Connector\Helper\Security as SecurityHelper;

/**
 * Lengow export
 */
class Export
{
    /* Export GET params */
    const PARAM_TOKEN = 'token';
    const PARAM_MODE = 'mode';
    const PARAM_FORMAT = 'format';
    const PARAM_STREAM = 'stream';
    const PARAM_OFFSET = 'offset';
    const PARAM_LIMIT = 'limit';
    const PARAM_TYPE = 'type';
    const PARAM_SELECTION = 'selection';
    const PARAM_OUT_OF_STOCK = 'out_of_stock';
    const PARAM_PRODUCT_IDS = 'product_ids';
    const PARAM_PRODUCT_TYPES = 'product_types';
    const PARAM_INACTIVE = 'inactive';
    const PARAM_STORE = 'store';
    const PARAM_STORE_ID = 'store_id';
    const PARAM_CODE = 'code';
    const PARAM_CURRENCY = 'currency';
    const PARAM_LANGUAGE = 'language';
    const PARAM_LOG_OUTPUT = 'log_output';
    const PARAM_UPDATE_EXPORT_DATE = 'update_export_date';
    const PARAM_GET_PARAMS = 'get_params';

    /* Legacy export GET params for old versions */
    const PARAM_LEGACY_LANGUAGE = 'locale';

    /**
     * @var string manual export type
     */
    const TYPE_MANUAL = 'manual';

    /**
     * @var string cron export type
     */
    const TYPE_CRON = 'cron';

    /**
     * @var string Magento cron export type
     */
    const TYPE_MAGENTO_CRON = 'magento cron';

    /**
     * @var StoreManagerInterface Magento store manager instance
     */
    protected $_storeManager;

    /**
     * @var DateTime Magento datetime instance
     */
    protected $_dateTime;

    /**
     * @var ScopeConfigInterface Magento scope config instance
     */
    protected $_scopeConfig;

    /**
     * @var ProductStatus Magento product status instance
     */
    protected $_productStatus;

    /**
     * @var ProductCollectionFactory Magento product collection factory
     */
    protected $_productCollectionFactory;

    /**
     * @var JsonHelper Magento json helper instance
     */
    protected $_jsonHelper;

    /**
     * @var WebsiteFactory Magento website factory instance
     */
    protected $_websiteFactory;

    /**
     * @var ConfigHelper Lengow config helper instance
     */
    protected $_configHelper;

    /**
     * @var DataHelper Lengow data helper instance
     */
    protected $_dataHelper;

    /**
     * @var LengowFeedFactory Lengow feed factory instance
     */
    protected $_feedFactory;

    /**
     * @var LengowProductFactory Lengow product factory instance
     */
    protected $_productFactory;

    /**
     * @var array all available params for export
     */
    protected $_exportParams = [
        self::PARAM_MODE,
        self::PARAM_FORMAT,
        self::PARAM_STREAM,
        self::PARAM_OFFSET,
        self::PARAM_LIMIT,
        self::PARAM_SELECTION,
        self::PARAM_OUT_OF_STOCK,
        self::PARAM_PRODUCT_IDS,
        self::PARAM_PRODUCT_TYPES,
        self::PARAM_INACTIVE,
        self::PARAM_STORE,
        self::PARAM_CODE,
        self::PARAM_CURRENCY,
        self::PARAM_LANGUAGE,
        self::PARAM_LOG_OUTPUT,
        self::PARAM_UPDATE_EXPORT_DATE,
        self::PARAM_GET_PARAMS,
    ];

    /**
     * @var array available formats for export
     */
    protected $_availableFormats = [
        LengowFeed::FORMAT_CSV,
        LengowFeed::FORMAT_YAML,
        LengowFeed::FORMAT_XML,
        LengowFeed::FORMAT_JSON,
    ];

    /**
     * @var array available formats for export
     */
    protected $_availableProductTypes = [
        'configurable',
        'simple',
        'downloadable',
        'grouped',
        'virtual',
    ];

    /**
     * @var array default fields for export
     */
    protected $_defaultFields = [
        'id' => 'id',
        'sku' => 'sku',
        'name' => 'name',
        'child_name' => 'child_name',
        'quantity' => 'quantity',
        'status' => 'status',
        'category' => 'category',
        'url' => 'url',
        'price_excl_tax' => 'price_excl_tax',
        'price_incl_tax' => 'price_incl_tax',
        'price_before_discount_excl_tax' => 'price_before_discount_excl_tax',
        'price_before_discount_incl_tax' => 'price_before_discount_incl_tax',
        'discount_amount' => 'discount_amount',
        'discount_percent' => 'discount_percent',
        'discount_start_date' => 'discount_start_date',
        'discount_end_date' => 'discount_end_date',
        'shipping_method' => 'shipping_method',
        'shipping_cost' => 'shipping_cost',
        'currency' => 'currency',
        'image_default' => 'image_default',
        'image_url_1' => 'image_url_1',
        'image_url_2' => 'image_url_2',
        'image_url_3' => 'image_url_3',
        'image_url_4' => 'image_url_4',
        'image_url_5' => 'image_url_5',
        'image_url_6' => 'image_url_6',
        'image_url_7' => 'image_url_7',
        'image_url_8' => 'image_url_8',
        'image_url_9' => 'image_url_9',
        'image_url_10' => 'image_url_10',
        'type' => 'type',
        'parent_id' => 'parent_id',
        'variation' => 'variation',
        'language' => 'language',
        'description' => 'description',
        'description_html' => 'description_html',
        'description_short' => 'description_short',
        'description_short_html' => 'description_short_html',
    ];

    /**
     * @var StoreInterceptor Magento store instance
     */
    protected $_store;

    /**
     * @var integer Magento store id
     */
    protected $_storeId;

    /**
     * @var integer amount of products to export
     */
    protected $_limit;

    /**
     * @var integer offset of total product
     */
    protected $_offset;

    /**
     * @var string format to return
     */
    protected $_format;

    /**
     * @var boolean stream return
     */
    protected $_stream;

    /**
     * @var string currency iso code for conversion
     */
    protected $_currency;

    /**
     * @var boolean export Lengow selection
     */
    protected $_selection;

    /**
     * @var boolean export out of stock product
     */
    protected $_outOfStock;

    /**
     * @var boolean include active products
     */
    protected $_inactive;

    /**
     * @var boolean see log or not
     */
    protected $_logOutput;

    /**
     * @var array export product types
     */
    protected $_productTypes;

    /**
     * @var array product ids to be exported
     */
    protected $_productIds;

    /**
     * @var boolean update export date.
     */
    protected $_updateExportDate;

    /**
     * @var string export type (manual, cron or magento cron)
     */
    protected $_exportType;

    /**
     * @var SecurityHelper Lengow security helper instance
     */
    protected $securityHelper;

    /**
     * Constructor
     *
     * @param StoreManagerInterface $storeManager Magento store manager instance
     * @param DateTime $dateTime Magento datetime instance
     * @param ScopeConfigInterface $scopeConfig Magento scope config instance
     * @param ProductStatus $productStatus Magento product status instance
     * @param ProductCollectionFactory $productCollectionFactory
     * @param JsonHelper $jsonHelper Magento json helper instance
     * @param WebsiteFactory $websiteFactory Magento website factory instance
     * @param DataHelper $dataHelper Lengow data helper instance
     * @param ConfigHelper $configHelper Lengow config helper instance
     * @param LengowFeedFactory $feedFactory Lengow feed factory instance
     * @param LengowProductFactory $productFactory Lengow product factory instance
     * @param SecurityHelper $securityHelper Lengow security helper instance
     */
    public function __construct(
        StoreManagerInterface $storeManager,
        DateTime $dateTime,
        ScopeConfigInterface $scopeConfig,
        ProductStatus $productStatus,
        ProductCollectionFactory $productCollectionFactory,
        JsonHelper $jsonHelper,
        WebsiteFactory $websiteFactory,
        DataHelper $dataHelper,
        ConfigHelper $configHelper,
        LengowFeedFactory $feedFactory,
        LengowProductFactory $productFactory,
        SecurityHelper $securityHelper
    ) {
        $this->_storeManager = $storeManager;
        $this->_dateTime = $dateTime;
        $this->_scopeConfig = $scopeConfig;
        $this->_productStatus = $productStatus;
        $this->_productCollectionFactory = $productCollectionFactory;
        $this->_jsonHelper = $jsonHelper;
        $this->_websiteFactory = $websiteFactory;
        $this->_dataHelper = $dataHelper;
        $this->_configHelper = $configHelper;
        $this->_feedFactory = $feedFactory;
        $this->_productFactory = $productFactory;
        $this->securityHelper = $securityHelper;
    }

    /**
     * Init a new export
     *
     * @param array $params optional options for init
     * integer store_id     ID of store
     * integer limit        The number of product to be exported
     * integer offset       From what product export
     * string  mode         Export mode => size: display only exported products, total: display all products
     * string  format       Export Format (csv|yaml|xml|json)
     * string  type         Type of export (manual, cron or magento cron)
     * string  product_type Type(s) of product
     * string  currency     Currency for export
     * string  product_ids  Ids product to export
     * boolean inactive     Export disabled products (1) | Export only enabled products (0)
     * boolean out_of_stock Export product in stock and out stock (1) | Export Only in stock product (0)
     * boolean selection    Export selected product (1) | Export all products (0)
     * boolean stream       Display file when call script (1) | Save File (0)
     * boolean log_output   See logs (only when stream = 0) (1) | no logs (0)
     */
    public function init($params)
    {
        $this->_storeId = isset($params[self::PARAM_STORE_ID]) ? (int) $params[self::PARAM_STORE_ID] : 0;
        try {
            $this->_store = $this->_storeManager->getStore($this->_storeId);
        } catch (\Exception $e) {
            $this->_store = $this->_storeManager->getDefaultStoreView();
        }
        $this->_limit = isset($params[self::PARAM_LIMIT]) ? (int) $params[self::PARAM_LIMIT] : 0;
        $this->_offset = isset($params[self::PARAM_OFFSET]) ? (int) $params[self::PARAM_OFFSET] : 0;
        $this->_stream = isset($params[self::PARAM_STREAM])
            ? (bool) $params[self::PARAM_STREAM]
            : !(bool) $this->_configHelper->get(ConfigHelper::EXPORT_FILE_ENABLED, $this->_storeId);
        $this->_selection = isset($params[self::PARAM_SELECTION])
            ? (bool) $params[self::PARAM_SELECTION]
            : (bool) $this->_configHelper->get(ConfigHelper::SELECTION_ENABLED, $this->_storeId);
        $this->_inactive = isset($params[self::PARAM_INACTIVE])
            ? (bool) $params[self::PARAM_INACTIVE]
            : (bool) $this->_configHelper->get(ConfigHelper::INACTIVE_ENABLED, $this->_storeId);
        $this->_outOfStock = isset($params[self::PARAM_OUT_OF_STOCK]) ? $params[self::PARAM_OUT_OF_STOCK] : true;
        $this->_updateExportDate = !isset($params[self::PARAM_UPDATE_EXPORT_DATE])
            || $params[self::PARAM_UPDATE_EXPORT_DATE];
        $this->_format = $this->_setFormat(
            isset($params[self::PARAM_FORMAT]) ? $params[self::PARAM_FORMAT] : LengowFeed::FORMAT_CSV
        );
        $this->_productIds = $this->_setProductIds(
            isset($params[self::PARAM_PRODUCT_IDS]) ? $params[self::PARAM_PRODUCT_IDS] : false
        );
        $this->_productTypes = $this->_setProductTypes(
            isset($params[self::PARAM_PRODUCT_TYPES]) ? $params[self::PARAM_PRODUCT_TYPES] : false
        );
        $this->_logOutput = $this->_setLogOutput(
            isset($params[self::PARAM_LOG_OUTPUT]) ? $params[self::PARAM_LOG_OUTPUT] : true
        );
        $this->_currency = $this->_setCurrency(
            isset($params[self::PARAM_CURRENCY]) ? $params[self::PARAM_CURRENCY] : false
        );
        $this->_exportType = $this->_setType(
            isset($params[self::PARAM_TYPE]) ? $params[self::PARAM_TYPE] : false
        );
    }

    /**
     * Get total available products
     *
     * @return integer
     **/
    public function getTotalProduct()
    {
        $productCollection = $this->_productCollectionFactory->create()
            ->setStoreId($this->_storeId)
            ->addStoreFilter($this->_storeId)
            ->addAttributeToFilter('type_id', ['nlike' => 'bundle']);
        return $productCollection->getSize();
    }

    /**
     * Get total exported products
     *
     * @return integer
     **/
    public function getTotalExportProduct()
    {
        $productCollection = $this->_getQuery();
        return $productCollection->getSize();
    }

    /**
     * Execute the export
     **/
    public function exec()
    {
        try {
            // start timer
            $timeStart = $this->_microtimeFloat();
            // clean logs
            $this->_dataHelper->cleanLog();
            $this->_dataHelper->log(
                DataHelper::CODE_EXPORT,
                $this->_dataHelper->setLogMessage('## start %1 export ##', [$this->_exportType]),
                $this->_logOutput
            );
            $this->_dataHelper->log(
                DataHelper::CODE_EXPORT,
                $this->_dataHelper->setLogMessage(
                    'start export in store %1 (%2)',
                    [
                        $this->_store->getName(),
                        $this->_storeId,
                    ]
                ),
                $this->_logOutput
            );
            // get fields to export
            $fields = $this->_getFields();
            // get products to be exported
            $productCollection = $this->_getQuery();
            $products = $productCollection->getData();
            $this->_dataHelper->log(
                DataHelper::CODE_EXPORT,
                $this->_dataHelper->setLogMessage('%1 product(s) found', [count($products)]),
                $this->_logOutput
            );
            $this->_export($products, $fields);
            if ($this->_updateExportDate) {
                $this->_configHelper->set(
                    ConfigHelper::LAST_UPDATE_EXPORT,
                    $this->_dateTime->gmtTimestamp(),
                    $this->_storeId
                );
            }
            $timeEnd = $this->_microtimeFloat();
            $this->_dataHelper->log(
                DataHelper::CODE_EXPORT,
                $this->_dataHelper->setLogMessage('memory usage: %1 Mb', [round(memory_get_usage() / 1000000, 2)]),
                $this->_logOutput
            );
            $this->_dataHelper->log(
                DataHelper::CODE_EXPORT,
                $this->_dataHelper->setLogMessage('execution time: %1 seconds', [round($timeEnd - $timeStart, 4)]),
                $this->_logOutput
            );
            $this->_dataHelper->log(
                DataHelper::CODE_EXPORT,
                $this->_dataHelper->setLogMessage('## end %1 export ##', [$this->_exportType]),
                $this->_logOutput
            );
        } catch (LengowException $e) {
            $errorMessage = $e->getMessage();
        } catch (\Exception $e) {
            $errorMessage = 'Magento error: "' . $e->getMessage() . '" ' . $e->getFile() . ' | ' . $e->getLine();
        }
        if (isset($errorMessage)) {
            $decodedMessage = $this->_dataHelper->decodeLogMessage($errorMessage, false);
            $this->_dataHelper->log(
                DataHelper::CODE_EXPORT,
                $this->_dataHelper->setLogMessage('export failed - %1', [$decodedMessage]),
                $this->_logOutput
            );
        }
    }

    /**
     * Export products
     *
     * @param array $products list of products to be exported
     * @param array $fields list of fields to export
     *
     * @throws \Exception|LengowException
     */
    protected function _export($products, $fields)
    {
        $isFirst = true;
        $productCount = 0;
        // get modulo for export counter
        $productModulo = $this->_getProductModulo(count($products));
        // get the maximum of character for yaml format
        $maxCharacter = $this->_getMaxCharacterSize($fields);
        // init product to export
        $lengowProduct = $this->_productFactory->create();
        $lengowProduct->init([
            'store' => $this->_store,
            'currency' => $this->_currency,
            'parentFields' => $this->_configHelper->getParentSelectedAttributes($this->_storeId)
        ]);
        // init feed to export
        $feed = $this->_feedFactory->create();
        $feed->init(
            [
                'stream' => $this->_stream,
                'format' => $this->_format,
                'store_code' => $this->_store->getCode(),
            ]
        );
        $feed->write(LengowFeed::HEADER, $fields);
        foreach ($products as $product) {
            $productData = [];
            $lengowProduct->load(
                [
                    'product_id' => (int) $product['entity_id'],
                    'product_type' => $product['type_id'],
                ]
            );
            if (!$this->_inactive && !$lengowProduct->isEnableForExport()) {
                $lengowProduct->clean();
                continue;
            }
            foreach ($fields as $field) {
                if (isset($this->_defaultFields[$field])) {
                    $productData[$field] = $lengowProduct->getData($this->_defaultFields[$field]);
                } else {
                    $productData[$field] = $lengowProduct->getData($field);
                }
            }
            // write product data
            $feed->write(LengowFeed::BODY, $productData, $isFirst, $maxCharacter);
            $productCount++;
            $this->_setCounterLog($productModulo, $productCount);
            // clean data for next product
            $lengowProduct->clean();
            unset($productData);
            $isFirst = false;
        }
        $success = $feed->end();
        if (!$success) {
            throw new LengowException(
                $this->_dataHelper->setLogMessage('unable to access the folder %1', [$feed->getFolderPath()])
            );
        }
        // product counter
        $counters = $lengowProduct->getCounters();
        $this->_dataHelper->log(
            DataHelper::CODE_EXPORT,
            $this->_dataHelper->setLogMessage(
                '%1 product(s) exported, %2 simple product(s), %3 configurable product(s),
                %4 grouped product(s), %5 virtual product(s), %6 downloadable product(s)',
                [
                    $counters['total'],
                    $counters['simple_enable'],
                    $counters['configurable'],
                    $counters['grouped'],
                    $counters['virtual'],
                    $counters['downloadable'],
                ]
            ),
            $this->_logOutput
        );
        // warning for simple product associated with configurable products disabled
        if ($counters['simple_disabled'] > 0) {
            $this->_dataHelper->log(
                DataHelper::CODE_EXPORT,
                $this->_dataHelper->setLogMessage(
                    'WARNING! %1 simple product(s) associated with configurable products are disabled',
                    [$counters['simple_disabled']]
                ),
                $this->_logOutput
            );
        }
        // link generation
        if (!$this->_stream) {
            $feedUrl = $feed->getUrl();
            if ($feedUrl) {
                $this->_dataHelper->log(
                    DataHelper::CODE_EXPORT,
                    $this->_dataHelper->setLogMessage(
                        'the export for the store %1 (%2) generated the following file: %3',
                        [
                            $this->_store->getName(),
                            $this->_storeId,
                            $feedUrl,
                        ]
                    ),
                    $this->_logOutput
                );
            }
        }
        unset($lengowProduct, $feed);
    }

    /**
     * Get all export available parameters
     *
     * @return string
     */
    public function getExportParams()
    {
        $params = [];
        $availableStores = [];
        $availableCodes = [];
        $availableCurrencies = [];
        $availableLanguages = [];
        foreach ($this->_websiteFactory->create()->getCollection() as $website) {
            foreach ($website->getGroups() as $group) {
                $stores = $group->getStores();
                foreach ($stores as $store) {
                    $availableStores[] = $store->getId();
                    $availableCodes[] = $store->getCode();
                    $currencyCodes = $store->getAvailableCurrencyCodes();
                    foreach ($currencyCodes as $currencyCode) {
                        if (!in_array($currencyCode, $availableCurrencies, true)) {
                            $availableCurrencies[] = $currencyCode;
                        }
                    }
                    $storeLanguage = $this->_scopeConfig->getValue(
                        'general/locale/code',
                        ScopeInterface::SCOPE_STORE,
                        $store->getId()
                    );
                    if (!in_array($storeLanguage, $availableLanguages, true)) {
                        $availableLanguages[] = $storeLanguage;
                    }
                }
            }
        }
        foreach ($this->_exportParams as $param) {
            switch ($param) {
                case self::PARAM_MODE:
                    $authorizedValue = ['size', 'total'];
                    $type = 'string';
                    $example = 'size';
                    break;
                case self::PARAM_FORMAT:
                    $authorizedValue = $this->_availableFormats;
                    $type = 'string';
                    $example = LengowFeed::FORMAT_CSV;
                    break;
                case self::PARAM_STORE:
                    $authorizedValue = $availableStores;
                    $type = 'integer';
                    $example = 1;
                    break;
                case self::PARAM_CODE:
                    $authorizedValue = $availableCodes;
                    $type = 'string';
                    $example = 'french';
                    break;
                case self::PARAM_CURRENCY:
                    $authorizedValue = $availableCurrencies;
                    $type = 'string';
                    $example = 'EUR';
                    break;
                case self::PARAM_LANGUAGE:
                    $authorizedValue = $availableLanguages;
                    $type = 'string';
                    $example = 'fr_FR';
                    break;
                case self::PARAM_OFFSET:
                case self::PARAM_LIMIT:
                    $authorizedValue = 'all integers';
                    $type = 'integer';
                    $example = 100;
                    break;
                case self::PARAM_PRODUCT_IDS:
                    $authorizedValue = 'all integers';
                    $type = 'string';
                    $example = '101,108,215';
                    break;
                case self::PARAM_PRODUCT_TYPES:
                    $authorizedValue = $this->_availableProductTypes;
                    $type = 'string';
                    $example = 'configurable,simple,grouped';
                    break;
                default:
                    $authorizedValue = [0, 1];
                    $type = 'integer';
                    $example = 1;
                    break;
            }
            $params[$param] = [
                'authorized_values' => $authorizedValue,
                'type' => $type,
                'example' => $example,
            ];
        }
        return $this->_jsonHelper->jsonEncode($params);
    }

    /**
     * Get fields to export
     *
     * @return array
     */
    protected function _getFields()
    {
        $fields = [];
        foreach ($this->_defaultFields as $key => $defaultField) {
            $fields[] = $key;
        }
        $selectedAttributes = $this->_configHelper->getSelectedAttributes($this->_storeId);
        foreach ($selectedAttributes as $selectedAttribute) {
            if (!in_array($selectedAttribute, $fields, true)) {
                $fields[] = $selectedAttribute;
            }
        }
        if ($this->_configHelper->moduleIsEnabled('Magento_Inventory')
            && version_compare($this->securityHelper->getMagentoVersion(), '2.3.0', '>=')
        ) {
            $sources = $this->_configHelper->getAllSources();
            // if multi-stock
            if (count($sources) > 1) {
                foreach ($sources as $source) {
                    $fields[] = 'quantity_multistock_' . $source;
                }
            }
        }
        return $fields;
    }

    /**
     * Set format to export
     *
     * @param string $format export format
     *
     * @return string
     */
    protected function _setFormat($format)
    {
        return !in_array($format, $this->_availableFormats, true) ? LengowFeed::FORMAT_CSV : $format;
    }

    /**
     * Set product ids to export
     *
     * @param boolean|array $productIds product ids to export
     *
     * @return array
     */
    protected function _setProductIds($productIds)
    {
        $ids = [];
        if ($productIds) {
            $exportedIds = explode(',', $productIds);
            foreach ($exportedIds as $id) {
                if (is_numeric($id) && $id > 0) {
                    $ids[] = (int) $id;
                }
            }
        }
        return $ids;
    }

    /**
     * Set product types to export
     *
     * @param boolean|array $productTypes product types to export
     *
     * @return array
     */
    protected function _setProductTypes($productTypes)
    {
        $types = [];
        if ($productTypes) {
            $exportedTypes = explode(',', $productTypes);
            foreach ($exportedTypes as $type) {
                if (in_array($type, $this->_availableProductTypes, true)) {
                    $types[] = $type;
                }
            }
        }
        if (empty($types)) {
            $types = explode(',', $this->_configHelper->get(ConfigHelper::EXPORT_PRODUCT_TYPES, $this->_storeId));
        }
        return $types;
    }

    /**
     * Set Log output for export
     *
     * @param boolean $logOutput see logs or not
     *
     * @return boolean
     */
    protected function _setLogOutput($logOutput)
    {
        return $this->_stream ? false : $logOutput;
    }

    /**
     * Set currency for export
     *
     * @param boolean|string $currency see logs or not
     *
     * @return string
     */
    protected function _setCurrency($currency)
    {
        $availableCurrencies = $this->_store->getAvailableCurrencyCodes();
        if (!$currency || !in_array($currency, $availableCurrencies, true)) {
            $currency = $this->_store->getCurrentCurrencyCode();
        }
        return $currency;
    }

    /**
     * Set export type
     *
     * @param boolean|string $type export type (manual, cron or magento cron)
     *
     * @return string
     */
    protected function _setType($type)
    {
        if (!$type) {
            $type = $this->_updateExportDate ? self::TYPE_CRON : self::TYPE_MANUAL;
        }
        return $type;
    }

    /**
     * Set the product counter log
     *
     * @param integer $productModulo product modulo
     * @param integer $productCount product counter
     *
     */
    protected function _setCounterLog($productModulo, $productCount)
    {
        $logMessage = $this->_dataHelper->setLogMessage('%1 product(s) exported', [$productCount]);
        // save 10 logs maximum in database
        if ($productCount % $productModulo === 0) {
            $this->_dataHelper->log(DataHelper::CODE_EXPORT, $logMessage);
        }
        if (!$this->_stream && $this->_logOutput) {
            if ($productCount % 50 === 0) {
                $countMessage = $this->_dataHelper->decodeLogMessage($logMessage, false);
                print_r('[Export] ' . $countMessage . '<br />');
            }
            flush();
        }
    }

    /**
     * Get product modulo for counter log
     *
     * @param integer $productTotal number of product to export
     *
     * @return integer
     */
    protected function _getProductModulo($productTotal)
    {
        $productModulo = (int)($productTotal / 10);
        return $productModulo < 50 ? 50 : $productModulo;
    }

    /**
     * Get max character size for yaml format
     *
     * @param array $fields list of fields to export
     *
     * @return integer
     */
    protected function _getMaxCharacterSize($fields)
    {
        $maxCharacter = 0;
        foreach ($fields as $field) {
            if (strlen($field) > $maxCharacter) {
                $maxCharacter = strlen($field);
            }
        }
        return $maxCharacter;
    }

    /**
     * Get microtime float
     */
    protected function _microtimeFloat()
    {
        list($usec, $sec) = explode(' ', microtime());
        return ((float) $usec + (float) $sec);
    }

    /**
     * Get products collection for export
     *
     * @return ProductCollection
     */
    protected function _getQuery()
    {
        // export only specific products types for one store
        $productCollection = $this->_productCollectionFactory->create()
            ->setStoreId($this->_storeId)
            ->addStoreFilter($this->_storeId)
            ->addAttributeToFilter('type_id', ['in' => $this->_productTypes]);
        // export only enabled products
        if (!$this->_inactive) {
            $productCollection->addAttributeToFilter('status', ['in' => $this->_productStatus->getVisibleStatusIds()]);
        }
        // export only selected products
        if ($this->_selection) {
            $productCollection->addAttributeToFilter('lengow_product', 1, 'left');
        }
        // export out of stock products
        if (!$this->_outOfStock) {
            try {
                $config = (int) $this->_scopeConfig->isSetFlag(CatalogInventoryConfiguration::XML_PATH_MANAGE_STOCK);
                $condition = '({{table}}.`is_in_stock` = 1) '
                    . ' OR IF({{table}}.`use_config_manage_stock` = 1, ' . $config . ', {{table}}.`manage_stock`) = 0';
                $productCollection->joinTable(
                    'cataloginventory_stock_item',
                    'product_id=entity_id',
                    ['qty' => 'qty', 'is_in_stock' => 'is_in_stock'],
                    $condition
                );
            } catch (\Exception $e) {
                $this->_dataHelper->log(
                    DataHelper::CODE_EXPORT,
                    $this->_dataHelper->setLogMessage(
                        'the junction with the %1 table did not work',
                        ['cataloginventory_stock_item']
                    ),
                    $this->_logOutput
                );
            }
        }
        // export specific products with id
        if (!empty($this->_productIds)) {
            $productCollection->addAttributeToFilter('entity_id', ['in' => $this->_productIds]);
        }
        // export with limit & offset
        if ($this->_limit > 0) {
            if ($this->_offset > 0) {
                $productCollection->getSelect()->limit($this->_limit, $this->_offset);
            } else {
                $productCollection->getSelect()->limit($this->_limit);
            }
        }
        return $productCollection;
    }
}
